/**
 * ProcessApplet - an applet to be used in ASPIRE context.
 */
package processView;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Iterator;
import java.util.Observable;
import java.util.Observer;
import java.util.Scanner;
import javax.swing.AbstractAction;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JApplet;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JToolBar;
import javax.swing.SwingUtilities;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.undo.UndoManager;
import javax.swing.undo.UndoableEdit;
import javax.swing.undo.UndoableEditSupport;
import org.jdesktop.swinghelper.layer.JXLayer;
import org.jdesktop.swinghelper.layer.painter.AbstractPainter;
import org.jdesktop.swinghelper.layer.painter.Painter;
import processView.FeedbackMessage;
import processView.ASPIREServerConnection;
import processModel.AdiabaticTransition;
import processModel.Formula;
import processModel.IsobaricTransition;
import processModel.IsochoricTransition;
import processModel.IsothermalTransition;
import processModel.Process;
import processModel.SpecialValue;
import processModel.State;
import processModel.Transition;
import processModel.TransitionUnknown;
import processModel.Unknown;
import utilities.*;
import java.util.*;
import netscape.javascript.*;

/**
 * ProcessApplet is one of the ContceteObservers participating in our
 * implementation of the Observer Design Pattern. A ProcessApplet object
 * observes a Process object, changes its state and querries it. Since Observer
 * Design Pattern is closely related to the Model-View-Controller System
 * architecture, this class implements the Contreller component of the applet.
 * ProcessDiagram class implements View component. Process class implements
 * Model component.
 * 
 * Undo/redo is implemented the Java way, namely through Command pattern and
 * UndoableEdit interface. The detailed description and example can be found
 * here: http://www.javaworld.com/javaworld/jw-06-1998/jw-06-undoredo.html.
 * 
 * @author Kon Zakharov, k.zakharov@cosc.canterbury.ac.nz
 * 
 */
public class ProcessApplet extends JApplet implements Observer {
	public static final boolean DEBUG = true;
	/**
	 * Required and autogenerated by Eclipse.
	 */
	private static final long serialVersionUID = -2720590184167392069L;
	private static final String appletVersion = "0.1";
	/**
	 * Name of the applet. To be used in output to identify the source.
	 */
	private static final String appletName = "Process Applet";
	/**
	 * Title to be shown at the top of the applet when displayed in the browser.
	 */
	// private static String appletTitle = "";
	/**
	 * Location of images required by the applet. Configurable through applet
	 * parameters. Default value is "images/".
	 */
	private static String imageDirectory = "images";
	private Color bgColor = Color.WHITE;
	private Container contentPane;
	protected UndoAction undoAction = null;
	protected RedoAction redoAction = null;
	protected JButton undoBtn_;
	protected JButton redoBtn_;
	protected JToolBar toolBar;

	/**
	 * The Subject participant of the Observer Design Pattern.
	 */
	private Process subject;
	private ProcessDiagram diagram;

	private FormulaWorkspace formulaWorkspace;
	private DataSummary dataSummary;
	private Frame pictureFrame = null;
	
	private ImageCache imageCache;

	protected URL base = null;
	/**
	 * Undo system elements
	 */
	protected UndoManager undoManager; // History list.
	protected UndoableEditSupport undoSupport; // Event support.

	private String userName;
	private String ontologyID;
	private ASPIREServerConnection serverConnection;
	private JSObject win;

	private boolean diagramDone = false;
	private boolean showFormulaPanel = true;

	private JXLayer<JComponent> layer;
	private Painter<JComponent> translucentPainter = new TranslucentPainter<JComponent>();

	public enum TutorStep {
		Diagram, SpecialValues, SelectingFormula, RearrangingFormula, SolvingUnknown
	}

	private TutorStep currentStep = TutorStep.Diagram;
	
	/*
	 * These are pairs are used to replace certain parts of the feedback messages before displaying them.
	 * The first of each pair is what to look for and the second is what to replace it with.
	 */
	private String[][] feedbackMessageMappings = 
		{{"You can correct your solution and press 'Submit' again, or try getting some more feedback.", "You can correct your solution and press 'Check Answer' again, or try getting some more feedback."},
		{"You can correct your solution and press 'Submit' again, or try getting some more feedback.", "You can correct your solution and press 'Check Answer' again, or try getting some more feedback."},
		{"Well done - you made only one mistake in the State concept.", "Well done - you made only one mistake in a state."},
		{"Almost there - a few mistakes though. One of them is in the State clause.", "Almost there - a few mistakes though. One of them is in a state."},
		{"Almost there - a few mistakes though. One of them is in the Isothermal compression clause.", "Almost there - a few mistakes though. One of them is in an Isothermal compression transition."},
		{"Well done - you made only one mistake in the Isothermal compression concept.", "Well done - you made only one mistake in an Isothermal compression transition."},
		{"Almost there - a few mistakes though. One of them is in the Isothermal expansion clause.", "Almost there - a few mistakes though. One of them is in an Isothermal expansion transition."},
		{"Well done - you made only one mistake in the Isothermal expansion concept.", "Well done - you made only one mistake in an Isothermal expansion transition."},
		{"Almost there - a few mistakes though. One of them is in the Adiabatic compression clause.", "Almost there - a few mistakes though. One of them is in an Adiabatic compression transition."},
		{"Well done - you made only one mistake in the Adiabatic compression concept.", "Well done - you made only one mistake in an Adiabatic compression transition."},
		{"Almost there - a few mistakes though. One of them is in the Adiabatic expansion clause.", "Almost there - a few mistakes though. One of them is in an Adiabatic expansion transition."},
		{"Well done - you made only one mistake in the Adiabatic expansion concept.", "Well done - you made only one mistake in an Adiabatic expansion transition."},
		{"Almost there - a few mistakes though. One of them is in the Isobaric compression clause.", "Almost there - a few mistakes though. One of them is in an Isobaric compression transition."},
		{"Well done - you made only one mistake in the Isobaric compression concept.", "Well done - you made only one mistake in an Isobaric compression transition."},
		{"Almost there - a few mistakes though. One of them is in the Isobaric expansion clause.", "Almost there - a few mistakes though. One of them is in an Isobaric expansion transition."},
		{"Well done - you made only one mistake in the Isobaric expansion concept.", "Well done - you made only one mistake in an Isobaric expansion transition."},
		{"Almost there - a few mistakes though. One of them is in the Isochoric pdecrease clause.", "Almost there - a few mistakes though. One of them is in an Isochoric pressure decrease transition."},
		{"Well done - you made only one mistake in the Isochoric pdecrease concept.", "Well done - you made only one mistake in an Isochoric pressure decrease transition."},
		{"Almost there - a few mistakes though. One of them is in the Isochoric pincrease clause.", "Almost there - a few mistakes though. One of them is in an Isochoric pressure increase transition."},
		{"Well done - you made only one mistake in the Isochoric pincrease concept.", "Well done - you made only one mistake in an Isochoric pressure increase transition."},
		{"Well done - you made only one mistake in the SpecialValue concept.", "Well done - you made only one mistake in a special value."},
		{"Almost there - a few mistakes though. One of them is in the SpecialValue clause.", "Almost there - a few mistakes though. One of them is in a special value."},
		{"Well done - you made only one mistake in the T-Unknown concept.", "Well done - you made only one mistake in the formula."},
		{"Almost there - a few mistakes though. One of them is in the T-Unknown clause.", "Almost there - a few mistakes though. One of them is in the formula."},
		{"Well done - you made only one mistake in the V-Unknown concept.", "Well done - you made only one mistake in the formula."},
		{"Almost there - a few mistakes though. One of them is in the V-Unknown clause.", "Almost there - a few mistakes though. One of them is in the formula."},
		{"Well done - you made only one mistake in the P-Unknown concept.", "Well done - you made only one mistake in the formula."},
		{"Almost there - a few mistakes though. One of them is in the P-Unknown clause.", "Almost there - a few mistakes though. One of them is in the formula."},
		{"Well done - you made only one mistake in the U-Unknown concept.", "Well done - you made only one mistake in the formula."},
		{"Almost there - a few mistakes though. One of them is in the U-Unknown clause.", "Almost there - a few mistakes though. One of them is in the formula."},
		{"Well done - you made only one mistake in the Q-Unknown concept.", "Well done - you made only one mistake in the formula."},
		{"Almost there - a few mistakes though. One of them is in the Q-Unknown clause.", "Almost there - a few mistakes though. One of them is in the formula."},
		{"Well done - you made only one mistake in the W-Unknown concept.", "Well done - you made only one mistake in the formula."},
		{"Almost there - a few mistakes though. One of them is in the W-Unknown clause.", "Almost there - a few mistakes though. One of them is in the formula."}};

	/**
	 * Returns information about applet's parameters. Browsers can use this
	 * information to help the user set your applet's parameter values.
	 * 
	 * @see java.applet.Applet#getParameterInfo()
	 */
	public String[][] getParameterInfo() {
		String[][] info = {
				// Parameter name Kind of value Description
				{ "appletTitle", "String", "Applet Title." },
				{
						"imageDir",
						"String",
						"Images directory. Default value: \"" + imageDirectory
								+ "\"" },
				{ "bgColor", "int list", "R|G|B values list, 0 - 255 range." }, };
		return info;
	}

	public String getAppletInfo() {
		// TODO: The return string should also contain the descriptions of the
		// applet interface methods. For example, our ASPIRE applets are
		// controlled
		// by JavaScript in the host pages -- JavaScript tells the applets to
		// load
		// or save a solution in XML on the server. So the methods that can be
		// called
		// by JavaScript should be documented in this info string.
		return appletName + " (Veresion: " + appletVersion
				+ ", serialVersionUID: " + serialVersionUID
				+ "): an applet to be used in ASPIRE context.";
	}

	public String getUserName() {
		return userName;
	}

	public String getOntologyID() {
		return ontologyID;
	}

	/**
	 * Parses the applet parameters passed through applet tag. Calls
	 * createGUI().
	 */
	/*
	 * The init method is useful for one-time initialization that doesn't take
	 * very long. In general, the init method should contain the code that you
	 * would normally put into a constructor. The reason applets shouldn't
	 * usually have constructors is that an applet isn't guaranteed to have a
	 * full environment until its init method is called. For example, the Applet
	 * image loading methods simply don't work inside of an applet constructor.
	 * The init method, on the other hand, is a great place to call the image
	 * loading methods, since the methods return quickly.
	 * 
	 * @see java.applet.Applet#init()
	 */
	public void init() {
		undoAction = new UndoAction();
		redoAction = new RedoAction();
		
		userName = getParameter("userName");
		ontologyID = getParameter("ontologyID");

		javax.swing.UIManager.put("TextField.font", AppConstants.arial11);
		javax.swing.UIManager.put("Label.font", AppConstants.arial11);
		javax.swing.UIManager.put("ComboBox.font", AppConstants.arial11);
		javax.swing.UIManager.put("List.font", AppConstants.arial11);
		javax.swing.UIManager.put("Button.font", AppConstants.arial11);

		if (DEBUG) {
			System.out.println(appletName + ": initialising.");
			// System.out.println("codeBase: " + this.getCodeBase());
			// System.out.println("documentBase: " + this.getDocumentBase());
		}
		showStatus(appletName + ": initialising.");

		// get the base URL
		if (base == null) {
			try {
				base = getDocumentBase();
			} catch (Exception e) {
				System.err.println("Unabled to initialise base");
				e.printStackTrace();
			}
		}

		// Get the background color parameter.
		String bgColorString = getParameter("bgColor");
		if (bgColorString != null) {
			try {
				Scanner scanner = new Scanner(bgColorString)
						.useDelimiter("\\|");
				bgColor = new Color(scanner.nextInt(), scanner.nextInt(),
						scanner.nextInt());
			} catch (Exception e) {
				System.err.println(e);
			}
		} else {
			if (DEBUG) {
				System.out.println("Applet parameter \"bgColor\" not found"
						+ ".");
			}
		}

		// TODO: Need to write code for processing applet title parameter.

		// TODO: Need to write code for processing image directory parameter.

		// Execute a job on the event-dispatching thread:
		// Creating this applet's GUI.
		try {
			SwingUtilities.invokeAndWait(new Runnable() {
				public void run() {
					createGUI();
				}
			});
		} catch (Exception e) {
			System.err.println("createGUI didn't complete successfully: " + e);
		}

		// TODO: This line must be commented out in order for the applet to start from eclipse
		win = JSObject.getWindow(this);
		serverConnection = new ASPIREServerConnection(this);
	}

	public void showMessageDialog(String title, String content, int type,
			Component parent) {
		JLabel msg = new JLabel(content);
		msg.setFont(AppConstants.arial11);
		JOptionPane pane = new JOptionPane(msg, type,
				JOptionPane.DEFAULT_OPTION);
		JPanel buttonPanel = (JPanel) pane.getComponent(1);
		Component[] comps = buttonPanel.getComponents();
		for (int i = 0; i < comps.length; i++) {
			if (comps[i] instanceof JButton) {
				comps[i].setFont(AppConstants.arial11);
			}
		}
		JDialog dialog = pane.createDialog(parent, title);
		dialog.setVisible(true);
	}

	private void createGUI() {
		contentPane = this.getContentPane();
		contentPane.setBackground(bgColor);
		contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.Y_AXIS));
		// JLabel title = new JLabel(appletTitle, SwingConstants.CENTER);
		// Set the label's dimensions to make sure it doesn't float around.
		Dimension dim = new Dimension(Integer.MAX_VALUE, 32);
		// title.setPreferredSize(dim);
		// title.setMaximumSize(dim);
		// title.setMinimumSize(dim);
		// Make it look the way we want.
		// title.setFont(new Font("SansSerif", Font.BOLD, 16));
		// title.setForeground(Color.GRAY);
		// Position it to the left.
		// title.setAlignmentX(SwingConstants.LEFT);
		// title.setHorizontalAlignment(SwingConstants.LEFT);
		// Create a border, but this is just for validating the layout. Can be
		// removed later.
		// title.setBorder(BorderFactory.createLineBorder(Color.GRAY));

		// Create the Model
		subject = new Process(this);
		// Create the View.
		diagram = new ProcessDiagram(subject);
		diagram.setPreferredSize(new Dimension(400, 270));
		// Set the subject's observer.
		subject.addObserver(this);
		subject.addObserver(diagram);

		// Add this title panel along with other components.
		// contentPane.add(title);
		contentPane.add(createToolBar());
		contentPane.add(diagram);

		layer = new JXLayer<JComponent>(diagram);
		layer.setPainter(translucentPainter);
		contentPane.add(layer, BorderLayout.CENTER);

		dataSummary = new DataSummary(this, subject.getTransitions(), subject
				.getStates());
		contentPane.add(dataSummary);
		formulaWorkspace = new FormulaWorkspace(this);
		formulaWorkspace.setEnabled(false);
		contentPane.add(formulaWorkspace, BorderLayout.SOUTH);

		validate();
		// Initilize the undo.redo system

		undoManager = new UndoManager();
		undoSupport = new UndoableEditSupport();
		undoSupport.addUndoableEditListener(new UndoAdapter());
		refreshUndoRedo();
	}

	private JToolBar createToolBar() {
		toolBar = new JToolBar();
		Dimension dim = new Dimension(Integer.MAX_VALUE,
				AppConstants.BUTTON_DIM);
		toolBar.setPreferredSize(dim);
		toolBar.setMaximumSize(dim);
		toolBar.setMinimumSize(dim);
		toolBar.setFloatable(false);

		// Action is a Swing interface that makes the UI the central point of
		// control.
		// That is, an action can be added directly to a toolbar (resulting in a
		// new button),
		// or to a menu (resulting in a new menu item). When the action changes
		// one of its
		// properties (for example, becomes enabled or disabled), the UI
		// elements are
		// notified and change their state accordingly. For example, when
		// CutAction becomes
		// disabled, both the Cut toolbar button and the Cut menu item will be
		// disabled
		// as well. In any case, the Action interface encapsulates the
		// ActionListener
		// interface for handling ActionEvent events, and a description of the
		// action itself.
		// Button dimensions in the toolbar are determined by the size of the
		// corresponding icons.
		toolBar.add(new ButtonAction(ActionEnum.NEW_STATE));
		toolBar.add(new ButtonAction(ActionEnum.NEW_ISOCHORIC));
		toolBar.add(new ButtonAction(ActionEnum.NEW_ISOBARIC));
		toolBar.add(new ButtonAction(ActionEnum.NEW_ISOTHERMAL));
		toolBar.add(new ButtonAction(ActionEnum.NEW_ADIABATIC));
		toolBar.addSeparator();
		undoBtn_ = new JButton(getImageCache().getImageIcon(ActionEnum.UNDO.icon));
		undoBtn_.addActionListener(undoAction);
		toolBar.add(undoBtn_);
		redoBtn_ = new JButton(getImageCache().getImageIcon(ActionEnum.REDO.icon));
		redoBtn_.addActionListener(redoAction);
		toolBar.add(redoBtn_);

		JButton btnDone = new JButton("Diagram Completed");
		btnDone.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String hintLevel = getFeedbackLevel();
				if (!hintLevel.equals("full-solution")) {
					submitStudentSolution(hintLevel);
					dataSummary.update();
				}
			}
		});

		// TODO: Remove this! For testing purposes only!
		/*final JComboBox cmbInsertAnswer = new JComboBox(new String[] { "Ans.",
				"1", "2", "3", "4", "5", "6", "7" });
		cmbInsertAnswer.setSize(10, 12);
		cmbInsertAnswer.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (cmbInsertAnswer.getSelectedItem().toString().equals("1")) {
					StateElement state1 = new StateElement();
					state1.setCenter(new Point(100, 50));
					state1.label("1");
					state1.temperature("600");
					state1.temperatureUnit("K");
					state1.pressure("1000000");
					state1.pressureUnit("Pa");
					state1.setVolumeUnknown();
					subject.addState(state1);

					StateElement state2 = new StateElement();
					state2.setCenter(new Point(220, 170));
					state2.label("2");
					state2.pressure("100000");
					state2.pressureUnit("Pa");
					state2.setTemperatureUnknown();
					state2.setVolumeUnknown();
					subject.addState(state2);

					StateElement state3 = new StateElement();
					state3.setCenter(new Point(220, 100));
					state3.label("3");
					state3.temperature("600");
					state3.temperatureUnit("K");
					state3.setPressureUnknown();
					state3.setVolumeUnknown();
					subject.addState(state3);

					Transition transition = new AdiabaticTransition("adiabatic");
					transition.label("1 - 2");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state1.getHandles().get(2));
					transition.getHandles().get(1).connect(
							state2.getHandles().get(3));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsochoricTransition("isochoric");
					transition.label("2 - 3");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state2.getHandles().get(0));
					transition.getHandles().get(1).connect(
							state3.getHandles().get(2));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsothermalTransition("isothermal");
					transition.label("3 - 1");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state3.getHandles().get(3));
					transition.getHandles().get(1).connect(
							state1.getHandles().get(1));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);
					diagram.repaint();
				} else if (cmbInsertAnswer.getSelectedItem().toString().equals(
						"2")) {
					StateElement state1 = new StateElement();
					state1.setCenter(new Point(100, 170));
					state1.label("1");
					state1.volume("0.3");
					state1.volumeUnit("m3");
					state1.pressure("400000");
					state1.pressureUnit("Pa");
					state1.setTemperatureUnknown();
					subject.addState(state1);

					StateElement state2 = new StateElement();
					state2.setCenter(new Point(200, 170));
					state2.label("2");
					state2.volume("0.5");
					state2.volumeUnit("m3");
					state2.pressure("400000");
					state2.pressureUnit("Pa");
					state2.setTemperatureUnknown();
					subject.addState(state2);

					StateElement state3 = new StateElement();
					state3.setCenter(new Point(100, 120));
					state3.label("3");
					state3.volume("0.3");
					state3.volumeUnit("m3");
					state3.setPressureUnknown();
					state3.setTemperatureUnknown();
					subject.addState(state3);

					Transition transition = new IsobaricTransition("isobaric");
					transition.label("1 - 2");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state1.getHandles().get(1));
					transition.getHandles().get(1).connect(
							state2.getHandles().get(3));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsothermalTransition("isothermal");
					transition.label("2 - 3");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state2.getHandles().get(0));
					transition.getHandles().get(1).connect(
							state3.getHandles().get(1));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsochoricTransition("isochoric");
					transition.label("3 - 1");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state3.getHandles().get(2));
					transition.getHandles().get(1).connect(
							state1.getHandles().get(0));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);
					diagram.repaint();
				} else if (cmbInsertAnswer.getSelectedItem().toString().equals(
						"3")) {
					StateElement state1 = new StateElement();
					state1.setCenter(new Point(170, 150));
					state1.label("1");
					state1.temperature("300");
					state1.temperatureUnit("K");
					state1.pressure("500000");
					state1.pressureUnit("Pa");
					state1.setVolumeUnknown();
					subject.addState(state1);

					StateElement state2 = new StateElement();
					state2.setCenter(new Point(100, 100));
					state2.label("2");
					state2.pressure("1000000");
					state2.pressureUnit("Pa");
					state2.setTemperatureUnknown();
					state2.setVolumeUnknown();
					subject.addState(state2);

					StateElement state3 = new StateElement();
					state3.setCenter(new Point(100, 150));
					state3.label("3");
					state3.pressure("500000");
					state3.pressureUnit("Pa");
					state3.setTemperatureUnknown();
					state3.setVolumeUnknown();
					subject.addState(state3);

					Transition transition = new AdiabaticTransition("adiabatic");
					transition.label("1 - 2");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state1.getHandles().get(0));
					transition.getHandles().get(1).connect(
							state2.getHandles().get(1));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsochoricTransition("isochoric");
					transition.label("2 - 3");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state2.getHandles().get(2));
					transition.getHandles().get(1).connect(
							state3.getHandles().get(0));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsobaricTransition("isobaric");
					transition.label("3 - 1");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state3.getHandles().get(1));
					transition.getHandles().get(1).connect(
							state1.getHandles().get(3));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);
					diagram.repaint();
				} else if (cmbInsertAnswer.getSelectedItem().toString().equals(
						"4")) {
					StateElement state1 = new StateElement();
					state1.setCenter(new Point(230, 170));
					state1.label("1");
					state1.temperature("500");
					state1.temperatureUnit("K");
					state1.setPressureUnknown();
					state1.setVolumeUnknown();
					subject.addState(state1);

					StateElement state2 = new StateElement();
					state2.setCenter(new Point(70, 50));
					state2.label("2");
					state2.temperature("800");
					state2.temperatureUnit("K");
					state2.setPressureUnknown();
					state2.setVolumeUnknown();
					subject.addState(state2);

					StateElement state3 = new StateElement();
					state3.setCenter(new Point(200, 50));
					state3.label("3");
					state3.temperature("2000");
					state3.temperatureUnit("K");
					state3.setPressureUnknown();
					state3.setVolumeUnknown();
					subject.addState(state3);

					StateElement state4 = new StateElement();
					state4.setCenter(new Point(230, 120));
					state4.label("4");
					state4.temperature("1000");
					state4.temperatureUnit("K");
					state4.setPressureUnknown();
					state4.setVolumeUnknown();
					subject.addState(state4);

					Transition transition = new AdiabaticTransition("adiabatic");
					transition.label("1 - 2");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state1.getHandles().get(3));
					transition.getHandles().get(1).connect(
							state2.getHandles().get(2));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsobaricTransition("isobaric");
					transition.label("2 - 3");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state2.getHandles().get(1));
					transition.getHandles().get(1).connect(
							state3.getHandles().get(3));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new AdiabaticTransition("adiabatic");
					transition.label("3 - 4");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state3.getHandles().get(1));
					transition.getHandles().get(1).connect(
							state4.getHandles().get(3));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsochoricTransition("isochoric");
					transition.label("4 - 1");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state4.getHandles().get(2));
					transition.getHandles().get(1).connect(
							state1.getHandles().get(0));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);
					diagram.repaint();
				} else if (cmbInsertAnswer.getSelectedItem().toString().equals(
						"5")) {
					StateElement state1 = new StateElement();
					state1.setCenter(new Point(100, 20));
					state1.label("1");
					state1.volume("1");
					state1.volumeUnit("m3");
					state1.temperature("368");
					state1.temperatureUnit("K");
					state1.setPressureUnknown();
					subject.addState(state1);

					StateElement state2 = new StateElement();
					state2.setCenter(new Point(200, 100));
					state2.label("2");
					state2.volume("2");
					state2.volumeUnit("m3");
					state2.temperature("368");
					state2.temperatureUnit("K");
					state2.setPressureUnknown();
					subject.addState(state2);

					StateElement state3 = new StateElement();
					state3.setCenter(new Point(200, 150));
					state3.label("3");
					state3.volume("2");
					state3.volumeUnit("m3");
					state3.temperature("297");
					state3.temperatureUnit("K");
					state3.setPressureUnknown();
					subject.addState(state3);

					StateElement state4 = new StateElement();
					state4.setCenter(new Point(100, 130));
					state4.label("4");
					state4.volume("1");
					state4.volumeUnit("m3");
					state4.temperature("297");
					state4.temperatureUnit("K");
					state4.setPressureUnknown();
					subject.addState(state4);

					Transition transition = new IsothermalTransition(
							"isothermal");
					transition.label("1 - 2");
					transition.setBeginPoint(new Point(100, 130));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state1.getHandles().get(1));
					transition.getHandles().get(1).connect(
							state2.getHandles().get(3));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsochoricTransition("isochoric");
					transition.label("2 - 3");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state2.getHandles().get(2));
					transition.getHandles().get(1).connect(
							state3.getHandles().get(0));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsothermalTransition("isothermal");
					transition.label("3 - 4");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state3.getHandles().get(3));
					transition.getHandles().get(1).connect(
							state4.getHandles().get(1));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsochoricTransition("isochoric");
					transition.label("4 - 1");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state4.getHandles().get(0));
					transition.getHandles().get(1).connect(
							state1.getHandles().get(2));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);
					diagram.repaint();
				} else if (cmbInsertAnswer.getSelectedItem().toString().equals(
						"6")) {
					StateElement state1 = new StateElement();
					state1.setCenter(new Point(100, 100));
					state1.label("1");
					state1.volume("1");
					state1.volumeUnit("m3");
					state1.setTemperatureUnknown();
					state1.setPressureUnknown();
					subject.addState(state1);

					StateElement state2 = new StateElement();
					state2.setCenter(new Point(200, 100));
					state2.label("2");
					state2.volume("2");
					state2.volumeUnit("m3");
					state2.temperature("800");
					state2.temperatureUnit("K");
					state2.setPressureUnknown();
					subject.addState(state2);

					StateElement state3 = new StateElement();
					state3.setCenter(new Point(200, 170));
					state3.label("3");
					state3.volume("2");
					state3.volumeUnit("m3");
					state3.setTemperatureUnknown();
					state3.setPressureUnknown();
					subject.addState(state3);

					StateElement state4 = new StateElement();
					state4.setCenter(new Point(100, 170));
					state4.label("4");
					state4.temperature("300");
					state4.temperatureUnit("K");
					state4.volume("1");
					state4.volumeUnit("m3");
					state4.setPressureUnknown();
					subject.addState(state4);

					Transition transition = new IsobaricTransition("isobaric");
					transition.label("1 - 2");
					transition.setBeginPoint(new Point(100, 130));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state1.getHandles().get(1));
					transition.getHandles().get(1).connect(
							state2.getHandles().get(3));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsochoricTransition("isochoric");
					transition.label("2 - 3");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state2.getHandles().get(2));
					transition.getHandles().get(1).connect(
							state3.getHandles().get(0));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsobaricTransition("isobaric");
					transition.label("3 - 4");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state3.getHandles().get(3));
					transition.getHandles().get(1).connect(
							state4.getHandles().get(1));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsochoricTransition("isochoric");
					transition.label("4 - 1");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state4.getHandles().get(0));
					transition.getHandles().get(1).connect(
							state1.getHandles().get(2));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);
					diagram.repaint();
				} else if (cmbInsertAnswer.getSelectedItem().toString().equals(
						"7")) {
					StateElement state1 = new StateElement();
					state1.setCenter(new Point(100, 60));
					state1.label("1");
					state1.pressure("500000");
					state1.pressureUnit("Pa");
					state1.setTemperatureUnknown();
					state1.setVolumeUnknown();
					subject.addState(state1);

					StateElement state2 = new StateElement();
					state2.setCenter(new Point(200, 60));
					state2.label("2");
					state2.pressure("500000");
					state2.pressureUnit("Pa");
					state2.setTemperatureUnknown();
					state2.setVolumeUnknown();
					subject.addState(state2);

					StateElement state3 = new StateElement();
					state3.setCenter(new Point(200, 150));
					state3.label("3");
					state3.pressure("100000");
					state3.pressureUnit("Pa");
					state3.temperature("284");
					state3.temperatureUnit("K");
					state3.setVolumeUnknown();
					subject.addState(state3);

					Transition transition = new IsobaricTransition("isobaric");
					transition.label("1 - 2");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state1.getHandles().get(1));
					transition.getHandles().get(1).connect(
							state2.getHandles().get(3));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new IsochoricTransition("isochoric");
					transition.label("2 - 3");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state2.getHandles().get(2));
					transition.getHandles().get(1).connect(
							state3.getHandles().get(0));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);

					transition = new AdiabaticTransition("adiabatic");
					transition.label("3 - 1");
					transition.setBeginPoint(new Point(100, 170));
					transition.setEndPoint(new Point(200, 170));
					transition.getHandles().get(0).connect(
							state3.getHandles().get(3));
					transition.getHandles().get(1).connect(
							state1.getHandles().get(2));
					transition.setDeltaUnknown();
					transition.setHeatUnknown();
					transition.setWorkUnknown();
					subject.addTransition(transition);
					diagram.repaint();
				}
			}
		});*/

		toolBar.addSeparator();
		toolBar.add(new ButtonAction(ActionEnum.DELETE));
		toolBar.add(new ButtonAction(ActionEnum.CLEAR));
		toolBar.add(btnDone);
		//cmbInsertAnswer.setMaximumSize(new Dimension(50, 24));
		//cmbInsertAnswer.setMinimumSize(new Dimension(40, 24));
		toolBar.addSeparator();
		// TODO: Change back after evaluation
		//toolBar.add(cmbInsertAnswer);
		return toolBar;
	}

	private String getFeedbackLevel() {
		return win.call("getFeedbackLevel", null).toString();
	}

	/*
	 * Every applet that does something after initialisation (except in direct
	 * response to user actions) must override the start method. The start
	 * method either performs the applet's work or (more likely) starts up one
	 * or more threads to perform the work.
	 * 
	 * @see java.applet.Applet#start()
	 */
	public void start() {
		if (DEBUG) {
			System.out.println(appletName + ": starting.");
		}
		System.out.println("Starting Control Applet");
		showStatus(appletName + ": starting.");
	}

	/*
	 * Most applets that override start should also override the stop method.
	 * The stop method should suspend the applet's execution, so that it doesn't
	 * take up system resources when the user isn't viewing the applet's page.
	 * For example, an applet that displays animation should stop trying to draw
	 * the animation when the user isn't looking at it.
	 * 
	 * @see java.applet.Applet#stop()
	 */
	public void stop() {
		if (DEBUG) {
			System.out.println(appletName + ": stopping.");
		}
		showStatus(appletName + ": stopping.");
	}

	/*
	 * Many applets don't need to override the destroy method, since their stop
	 * method (which is called before destroy) does everything necessary to shut
	 * down the applet's execution. However, destroy is available for applets
	 * that need to release additional resources.
	 * 
	 * @see java.applet.Applet#destroy()
	 */
	public void destroy() {
		if (DEBUG) {
			System.out.println(appletName + ": destroying.");
		}
		showStatus(appletName + ": destroying.");
		// This is where we need to detach subject's observers.
		// TODO: One thing that needs to be clarified is whether
		// our observers need to hang onto their subjects.
		// I think at the moment we don't need to hang on them.
		// The Observer interface seems to indicate we just need to
		// pass the subjects in update calls.
		subject.deleteObservers();
	}

	/**
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	public void update(Observable o, Object arg) {
		System.out.println("update in ProcessApplet: " + arg + " from " + o);
		Process process = ((Process) o);
		UndoableEdit edit = process.getLastEdit();
		process.lastOperation(null);
		if (edit != null) {
			undoSupport.beginUpdate();
			undoSupport.postEdit(edit); // notify the listeners
			undoSupport.endUpdate();
			System.out
					.println("In processApplet.update, PROCESS has changed! - edit is: "
							+ edit);
		}
		refreshUndoRedo();
	}

	public void refreshUndoRedo() {

		// refresh undo
		undoBtn_.setText(undoManager.getUndoPresentationName());
		undoBtn_.setEnabled(undoManager.canUndo());
		// refresh redo
		redoBtn_.setText(undoManager.getRedoPresentationName());
		redoBtn_.setEnabled(undoManager.canRedo());
	}

	public Process getProcess() {
		return subject;
	}

	public FormulaWorkspace getWorkspace() {
		return formulaWorkspace;
	}
	
	public boolean submitStudentSolution()
	{
		return submitStudentSolution(getFeedbackLevel());
	}
	
	public boolean submitStudentSolution(String hintLevel) {
		// System.out.println("Enter Thermo Diagram");
		// If no errors in force diagram
		String studentSolution = getStudentSolution();
		System.out.println("----------------------------");
		System.out.println("Here's what's being posted (" + currentStep + "):");
		System.out.println(studentSolution);
		System.out.println("----------------------------");
		FeedbackMessage msg = serverConnection.submitSolution(solutionForPage(
				studentSolution, getPageNumber()), hintLevel);
		if (msg.isSolutionCorrect()) {
			
			// Set custom message so it won't say go to next page ..
			if (currentStep.equals(TutorStep.Diagram)) {
				msg.setFeedbackText("<p>Congratulations, the thermodynamics diagram is complete.</p>"
						+ "<p>Now use the buttons on the data summary to select values for the five constants.</p>");
			}
			else if (currentStep.equals(TutorStep.SpecialValues)) {
				msg.setFeedbackText("<p>Congratulations, the five constants are correct.</p>"
						+ "<p>Now use the formula panel below to calculate the unknowns.</p>"
						+ "<p>Begin by selecting the unknown to calculate then select a formula to use from "
						+ "one of the groups.</p>");
			}
			else if (currentStep.equals(TutorStep.SelectingFormula)){
				if (formulaWorkspace.getCurrentFormulaPanel().getSelectedFormula().mustRearrange()) {
					msg.setFeedbackText("<p>The formula you have selected is correct!</p>"
							+ "<p>You can rearrange the formula by clicking \"Rearrange\"</p>");
				}
				else if (formulaWorkspace.getCurrentFormulaPanel().getSelectedFormula().canRearrange()) {
					msg.setFeedbackText("<p>The formula you have selected is correct!</p>"
							+ "<p>You can rearrange the formula if necessary by clicking \"Rearrange\"</p>"
							+ "<p>Otherwise fill in the unknowns by right-clicking values in the table "
							+ "and click \"Evaluate\" when you are done.");
				}
				else {
					msg.setFeedbackText("<p>The formula you have selected is correct!</p>"
							+ "<p>Now fill in the unknowns by right-clicking values in the table.</p>"
							+ "<p>When you are done click \"Evaluate\"</p>");
					// Advance an extra step because the formula cannot be rearranged
					advanceToNextStep();
				}
			}
			else if (currentStep.equals(TutorStep.RearrangingFormula)){
				msg.setFeedbackText("<p>The rearranged formula you have selected is correct!</p>"
						+ "<p>Now fill in the unknowns by right-clicking values in the table</p>"
						+ "<p>When you are done click \"Evaluate\"</p>");
			}
			else if (currentStep.equals(TutorStep.SolvingUnknown)) {
				if (solvedAllTransitions()) {
					msg.setFeedbackText("<p>Congratulations, you have correctly calculated all unknowns.</p>"
							+ "<p>Please select another problem, or click \"Give Me a Problem\" from the top of the page.</p>");
				}
				else {
					msg.setFeedbackText("<p>Well done, your answer is correct</p>"
							+ "<p>Click \"Calculate another unknown\" to continue.</p>");
					this.formulaWorkspace.enableAddButton();
					this.formulaWorkspace.getCurrentFormulaPanel().setEnabled(false);
				}
			}
			msg.setNextFeedbackLevel("pos-neg");
			advanceToNextStep();
		}
		else {
			long time = System.currentTimeMillis();
			filterFeedbackMessage(msg);
			System.out.println("Filtering Took: " + (System.currentTimeMillis()-time) + " milliseconds");
		}
		displayASPIREFeedback(msg);
		return msg.isSolutionCorrect();
	}
	
	private void filterFeedbackMessage(FeedbackMessage msg)
	{
		for (String[] pair : feedbackMessageMappings)
		{
			msg.setFeedbackText(msg.getFeedbackText().replace(pair[0], pair[1]));
		}
	}
	
	private boolean solvedAllTransitions()
	{
		for (Unknown unknown : subject.getUnknowns()) {
			if (unknown instanceof TransitionUnknown) {
				return false;
			}
		}
		return true;
	}
	
	private void advanceToNextStep()
	{
		if (currentStep.equals(TutorStep.Diagram)) {
			currentStep = TutorStep.SpecialValues;
			enableSpecialValuesPanel();
		}
		else if (currentStep.equals(TutorStep.SpecialValues)) {
			enableFormulaPanel();
			currentStep = TutorStep.SelectingFormula;
		}
		else if (currentStep.equals(TutorStep.SelectingFormula)) {
			currentStep = TutorStep.RearrangingFormula;
		}
		else if (currentStep.equals(TutorStep.RearrangingFormula)) {
			currentStep = TutorStep.SolvingUnknown;
		}
		else if (currentStep.equals(TutorStep.SolvingUnknown)) {
			currentStep = TutorStep.SelectingFormula;
		}
		System.out.println("Current Step:  " + currentStep);
	}
	
	/**
	 * Over-ride the current step.  This is necessary to allow the student
	 * to go back to previous steps while solving an unknown.
	 * @param newCurrentStep
	 */
	public void setCurrentStep(TutorStep newCurrentStep)
	{
		currentStep = newCurrentStep;
	}

	/**
	 * This method is called when the "Check Answer" button on the HTML page is
	 * clicked.
	 */
	public void checkAnswerButtonClicked(String hintLevel) {
		formulaWorkspace.evaluateCurrentEquation();
		submitStudentSolution(hintLevel);
		dataSummary.update();
	}
	
	private void enableSpecialValuesPanel()
	{
		layer.setLocked(true);
		for (int i=0;i<toolBar.getComponentCount(); i++)
		toolBar.getComponents()[i].setEnabled(false);
		this.diagram.disableDiagram();
		this.diagramDone = true;
		dataSummary.enableSpecialValuesPanel();
	}

	private void enableFormulaPanel()
	{
		dataSummary.setSpecialValuesComplete();
		dataSummary.update();
		if (this.showFormulaPanel) {
			// Activate formula workspace
			formulaWorkspace.setEnabled(true);
			// Add a formula panel by default
			formulaWorkspace.addFormulaPanel();
		} else {
			String args[] = { this.subject.getUnknownsXML() };
			win.call("setFormulaPanelUnknowns", args);
		}
		// Must enable submit button because it may have been disabled by JS
		// submit button
		// if solution is correct
		win.call("enableSubmitButton", null);
		win.call("enableFeedbackLevel", null);
		String args[] = { "2" };
		win.call("setPageNumber", args);
	}

	public void checkSpecialValueStep(DataSummary dataSummary, String value) {
		List<SpecialValue> specialValuesList = null;
		if (value.equals(SpecialValue.SPECIAL_VALUE_CV)) {
			specialValuesList = createSpecialValueCvList();
		} else if (value.equals(SpecialValue.SPECIAL_VALUE_CP)) {
			specialValuesList = createSpecialValueCpList();
		} else if (value.equals(SpecialValue.SPECIAL_VALUE_R)) {
			specialValuesList = createSpecialValueRList();
		} else if (value.equals(SpecialValue.SPECIAL_VALUE_GAMMA)) {
			specialValuesList = createSpecialValueGammaList();
		} else if (value.equals(SpecialValue.SPECIAL_VALUE_N)) {
			specialValuesList = createSpecialValueNList();
		}

		SpecialValueMenu menu = new SpecialValueMenu(this, dataSummary,
				pictureFrame, "Select special value", subject,
				specialValuesList);
		menu.pack();
		Point appLocation = this.formulaWorkspace.getLocationOnScreen();
		System.out.println("formulaWorkspace: " + formulaWorkspace.getLocationOnScreen());
		menu.setLocation(appLocation.x + formulaWorkspace.getWidth()/2 - menu.getWidth()/2, appLocation.y);
		menu.setVisible(true);
	}

	public void checkFormulaStepAnswer(FormulaPanel panel,
			String stepIteration, String step, String value) {
		formulaWorkspace.setLastStep(step);
		checkFormulaStep(panel, formulaWorkspace.toASPIRE_XML(stepIteration,
				step, value), step, value);
	}

	public void checkFormulaStep(FormulaPanel panel, String solution,
			String step, String value) {
		PictureMenu d = null;
		String solutionXML = solutionForPage(solution, 2);
		System.out.println("This is it: " + solutionXML);
		
		if (step.equals(FormulaWorkspace.FORMULA)) {
			System.out.println("HERE:  " + value);
			List<Formula> equationList = null;

			if (value.equals(FormulaPanel.EQN_ADIABATIC)) {
				equationList = createAdiabaticEquationList();
			} else if (value.equals(FormulaPanel.EQN_DELTA_U)) {
				equationList = createDeltaUEquationList();
			} else if (value.equals(FormulaPanel.EQN_FIRST_LAW)) {
				equationList = createFirstLawEquationList();
			} else if (value.equals(FormulaPanel.EQN_IDEAL_GAS_STATE)) {
				equationList = createIdealGasEquationList();
			} else if (value.equals(FormulaPanel.EQN_ISOTHERMAL_WORK)) {
				equationList = createIsothermalWorkEquationList();
			} else if (value.equals(FormulaPanel.EQN_SPECIAL_PROPERTIES)) {
				equationList = createSpecialPropertiesEquationList();
			} else if (value.equals(FormulaPanel.EQN_ISOBARIC_WORK)) {
				equationList = createIsobaricWorkEquationList();
			}
			d = new FormulaPictureMenu(this, panel, pictureFrame,
					"Select the correct Formula", subject, equationList,
					"fva");
		} else if (step.equals(FormulaWorkspace.REARRANGED_FORMULA)) {
			List<Formula> equationList = null;
			equationList = panel.getSelectedFormula()
					.getRearrangedEquations();
			d = new FormulaPictureMenu(this, panel, pictureFrame,
					"Select the correct Formula", subject, equationList,
					"fva");
		}
		
		if (d != null)
		{
			d.pack();
			if (d.getWidth() > 722) {
				d.setSize(722, (int)(d.getHeight()*1.65));
			}
			Point appLocation = this.formulaWorkspace.getLocationOnScreen();
			d.setLocation(appLocation.x + formulaWorkspace.getWidth()/2 - d.getWidth()/2, appLocation.y+30);
			d.setVisible(true);
		}

		if (!this.showFormulaPanel) {
			// Must call JS function that informs dll
			win.call("formulaStepCorrect", null);
		}
	}

	private String solutionForPage(String solution, int pageNo) {
		return "<student-solution page-number='" + pageNo + "'>\n" + solution
				+ "</student-solution>";
	}

	public void displayASPIREFeedback(FeedbackMessage msg) {
		String args[] = { msg.getFeedbackText(), msg.getSolutionCorrect(),
				msg.getNextFeedbackLevel() };
		win.call("displayFeedback", args);
	}

	public int getPageNumber() {
		int pageNumber = 0;
		if (currentStep.equals(TutorStep.Diagram)) {
			pageNumber = 1;
		} else if (currentStep.equals(TutorStep.SpecialValues)) {
			pageNumber = 2;
		} else if (currentStep.equals(TutorStep.SelectingFormula)) {
			pageNumber = 3;
		} else if (currentStep.equals(TutorStep.RearrangingFormula)) {
			pageNumber = 4;
		} else if (currentStep.equals(TutorStep.SolvingUnknown)) {
			pageNumber = 5;
		}
		System.out.println("Current page number:  " + pageNumber);
		return pageNumber;
	}

	public String getStudentSolution() {
		if (currentStep.equals(TutorStep.Diagram)) {
			return subject.diagramToASPIREXML();
		} else if (currentStep.equals(TutorStep.SpecialValues)) {
			return dataSummary.specialValuesToASPIREXML();
		} else if (currentStep.equals(TutorStep.SelectingFormula)) {
			return formulaWorkspace.solvingUnknownsToASPIREXML();
		} else if (currentStep.equals(TutorStep.RearrangingFormula)) {
			return formulaWorkspace.solvingUnknownsToASPIREXML();
		} else if (currentStep.equals(TutorStep.SolvingUnknown)) {
			return formulaWorkspace.solvingUnknownsToASPIREXML();
		}
		System.out.println("getStudentSolution returning null");
		return null;
	}

	/**
	 * Aspire calls this method to let the Applet know the solution was correct.
	 */
	public void isCorrect(boolean value) {
		System.out.println("Aspire notified applet, solution was correct: "
				+ value);
	}
	
	public ImageCache getImageCache()
	{
		if (imageCache == null) {
			imageCache = new ImageCache(this);
		}
		return imageCache;
	}

	/*public ImageIcon getImageIcon(String name) {

		Image image = getImage(base, imageDirectory + "/" + name);
		ImageIcon icon = null;
		try {
			icon = new ImageIcon(image);
		} catch (Exception ex) {
			System.err.println("Couldn't find file: " + base + " :: " + name);
		}
		return icon;*/
		/*
		 * Image img = null; // TODO: This code needs reviewing. sun.java.com
		 * has some material on loading images in applets. //
		 * http://java.sun.com/docs/books/tutorial/uiswing/components/icon.html
		 * URLClassLoader urlLoader =
		 * (URLClassLoader)this.getClass().getClassLoader(); URL fileLoc =
		 * urlLoader.findResource(name); // TODO: We need to have some code that
		 * sets the button image to some default // virtual icon or textual
		 * label. This can be done with an exception. Basically, // this method
		 * should throw an exception. //
		 * System.err.println(System.getProperty("java.class.path")); if(fileLoc
		 * != null) { System.err.println("URL OK: " + fileLoc); img =
		 * this.getToolkit().createImage(fileLoc); MediaTracker tracker = new
		 * MediaTracker(this); tracker.addImage(img, 0); try {
		 * tracker.waitForID(0); if (tracker.isErrorAny()) {
		 * System.out.println("Error loading image " + fileLoc); } } catch
		 * (Exception ex) { ex.printStackTrace(); } return new ImageIcon(img); }
		 * else { System.err.println("URL not valid: " + name); return null; }
		 */
		/*
		 * // This is another simple approach to loading images. URL url = null;
		 * try { url = new URL(getCodeBase() + imageDirectory + b.getImgName());
		 * System.err.println(url); } catch (Exception e) {
		 * System.err.println(e); } ImageIcon icon = new ImageIcon(url);
		 */
	//}


	class ButtonAction extends AbstractAction {
		ActionEnum actEnum;

		public ButtonAction(ActionEnum ae) {
			super();
			actEnum = ae;
			// This is a more robust threaded approach recommended at
			//http://java.sun.com/docs/books/tutorial/uiswing/components/applet.
			// html
			// http://java.sun.com/docs/books/tutorial/uiswing/components/icon.
			// html
			// but it is not working very well. So much for a robust approach!
			// Possibly, it will work OK when the applet is retreived from the
			// server,
			// but for our development and debugging it doesnt' look like we can
			// use it.
			// Anyway, try it later, when the applet is on the server.
			// ImageIcon icon = loadImageIcon(iconPath, actEnum.description);

			// This is yet another method I wrote a while ago.
			ImageIcon icon = getImageCache().getImageIcon(ae.icon);
			if (icon == null) {
				super.putValue(NAME, ae.name);
			} else {
				super.putValue(SMALL_ICON, icon);
			}
			// SHORT_DESCRIPTION should be used as the tooltip.
			super.putValue(SHORT_DESCRIPTION, ae.description);
		}

		/**
		 * Required and autogenerated by Eclipse.
		 */
		private static final long serialVersionUID = 178761300092475793L;

		public void actionPerformed(ActionEvent e) {
			// perform the operation
			((Process) subject).setState(actEnum);
		}

	}

	/**
	 * An undo/redo adpater. The adpater is notified when an undo edit
	 * occur(e.g. add or remove from the list) The adptor extract the edit from
	 * the event, add it to the UndoManager, and refresh the GUI
	 */
	private class UndoAdapter implements UndoableEditListener {
		public void undoableEditHappened(UndoableEditEvent evt) {
			UndoableEdit edit = evt.getEdit();
			System.out
					.println("In processApplet.undoableEditHappened(), edit = "
							+ edit);
			if (edit != null)
				undoManager.addEdit(edit);
			refreshUndoRedo();
		}
	}

	/**
	 * Undo action.
	 */
	private class UndoAction extends AbstractAction {
		private static final long serialVersionUID = -3703995935350546187L;

		UndoAction() {
			super(ActionEnum.UNDO.name());
			// This is a more robust threaded approach recommended at
			//http://java.sun.com/docs/books/tutorial/uiswing/components/applet.
			// html
			// http://java.sun.com/docs/books/tutorial/uiswing/components/icon.
			// html
			// but it is not working very well. So much for a robust approach!
			// Possibly, it will work OK when the applet is retreived from the
			// server,
			// but for our development and debugging it doesnt' look like we can
			// use it.
			// Anyway, try it later, when the applet is on the server.
			// ImageIcon icon = loadImageIcon(iconPath, actEnum.description);
			// This is yet another method I wrote a whire ago.
			ImageIcon icon = getImageCache().getImageIcon(ActionEnum.UNDO.icon);
			if (icon == null) {
				super.putValue(NAME, "?");
			} else {
				super.putValue(SMALL_ICON, icon);
			}
			// SHORT_DESCRIPTION should be used as the tooltip.
			super.putValue(SHORT_DESCRIPTION, ActionEnum.UNDO.description);
		}

		public void actionPerformed(ActionEvent evt) {
			undoManager.undo();
			((Process) subject).setState(ActionEnum.UNDO);
			repaint();
			refreshUndoRedo();
		}
	}

	/**
	 * Redo action.
	 */
	private class RedoAction extends AbstractAction {
		private static final long serialVersionUID = -8693489515307921889L;

		RedoAction() {
			super(ActionEnum.REDO.name());
			// This is a more robust threaded approach recommended at
			//http://java.sun.com/docs/books/tutorial/uiswing/components/applet.
			// html
			// http://java.sun.com/docs/books/tutorial/uiswing/components/icon.
			// html
			// but it is not working very well. So much for a robust approach!
			// Possibly, it will work OK when the applet is retreived from the
			// server,
			// but for our development and debugging it doesnt' look like we can
			// use it.
			// Anyway, try it later, when the applet is on the server.
			// ImageIcon icon = loadImageIcon(iconPath, actEnum.description);
			// This is yet another method I wrote a whire ago.
			ImageIcon icon = getImageCache().getImageIcon(ActionEnum.REDO.icon);
			if (icon == null) {
				super.putValue(NAME, "?");
			} else {
				super.putValue(SMALL_ICON, icon);
			}
			// SHORT_DESCRIPTION should be used as the tooltip.
			super.putValue(SHORT_DESCRIPTION, ActionEnum.REDO.description);
		}

		public void actionPerformed(ActionEvent evt) {
			undoManager.redo();
			((Process) subject).setState(ActionEnum.REDO);
			repaint();
			refreshUndoRedo();
		}
	}
	
	public void postErrorMessage(String message)
	{
		FeedbackMessage msg = new FeedbackMessage();
		msg.setFeedbackText("<p>You have made a small error.</p>" + 
				"<p>" + message + "</p>" +
				"<p>You can correct your solution and press 'Submit' again, or try getting some more feedback.</p>" +
				"<p>Would you like to have another go?</p>");
		msg.setSolutionCorrect("f");
		displayASPIREFeedback(msg);
	}

	public void updateDataSummary() {
		dataSummary.update();
	}

	static class TranslucentPainter<V extends JComponent> extends
			AbstractPainter<V> {
		public void paint(Graphics2D g2, JXLayer<V> l) {
			l.paint(g2);
			if (l.isLocked()) {
				g2.setColor(new Color(255, 255, 255, 128));
				g2.fillRect(0, 0, l.getWidth(), l.getHeight());
			}
		}
	}

	/**
	 * private method that creates <code>equationList</code>, a list that
	 * contains all the Capital Investment Evaluation Technique Equations. It
	 * then shuffles the list, so that the order of appearance on the panel is
	 * different each time the student views the panel.
	 */
	private List<Formula> createFirstLawEquationList() {
		ArrayList<Formula> equationList = new ArrayList<Formula>(); //<Equation
																		// >
		Formula first_law_delta_q = new Formula("FirstLaw", "first_law_delta_q",
				new String[] { "Q = ", "\u0394U" }, true, false,
				base, "first_law_delta_q.gif");
		Formula first_law_delta_q_rearranged = new Formula("FirstLaw", 
				"first_law_delta_q_rearranged", new String[] {
						"\u0394U = ", "Q" }, true, false, base,
				"first_law_delta_q_rearranged.gif");
		first_law_delta_q.addRearrangedEquation(first_law_delta_q_rearranged);
		first_law_delta_q_rearranged.addRearrangedEquation(first_law_delta_q);
		first_law_delta_q_rearranged.setParentEquation(first_law_delta_q);

		Formula first_law_delta_qw = new Formula("FirstLaw", "first_law_delta_qw",
				new String[] { "\u0394U = ", "Q", "+", "W" }, true, false,
				base, "first_law_delta_qw.gif");
		Formula first_law_delta_qw_rearranged_q = new Formula("FirstLaw", 
				"first_law_delta_qw_rearranged_q",
				new String[] { "Q = ", "\u0394U", "-", "W" }, true, false,
				base, "first_law_delta_qw_rearranged_q.gif");
		Formula first_law_delta_qw_rearranged_w = new Formula("FirstLaw", 
				"first_law_delta_qw_rearranged_w",
				new String[] { "W = ", "\u0394U", "-", "Q" }, true, false,
				base, "first_law_delta_qw_rearranged_w.gif");
		first_law_delta_qw
				.addRearrangedEquation(first_law_delta_qw_rearranged_q);
		first_law_delta_qw
				.addRearrangedEquation(first_law_delta_qw_rearranged_w);
		first_law_delta_qw_rearranged_q
				.addRearrangedEquation(first_law_delta_qw);
		first_law_delta_qw_rearranged_q
				.addRearrangedEquation(first_law_delta_qw_rearranged_w);
		first_law_delta_qw_rearranged_w
				.addRearrangedEquation(first_law_delta_qw);
		first_law_delta_qw_rearranged_w
				.addRearrangedEquation(first_law_delta_qw_rearranged_q);
		
		first_law_delta_qw_rearranged_q.setParentEquation(first_law_delta_qw);
		first_law_delta_qw_rearranged_w.setParentEquation(first_law_delta_qw);

		Formula first_law_delta_w = new Formula("FirstLaw", "first_law_delta_w",
				new String[] { "W = ", "\u0394U" }, true, false,
				base, "first_law_delta_w.gif");
		Formula first_law_delta_w_rearranged = new Formula("FirstLaw", 
				"first_law_delta_w_rearranged", new String[] {
						"\u0394U = ", "W" }, true, false, base,
				"first_law_delta_w_rearranged.gif");
		first_law_delta_w.addRearrangedEquation(first_law_delta_w_rearranged);
		first_law_delta_w_rearranged.addRearrangedEquation(first_law_delta_w);
		
		first_law_delta_w_rearranged.setParentEquation(first_law_delta_w);

		Formula first_law_q = new Formula("FirstLaw", "first_law_q",
				new String[] { "Q = ", "\u0394U" }, true, false, base,
				"first_law_q.gif");
		Formula first_law_q_rearranged = new Formula("FirstLaw", 
				"first_law_q_rearranged",
				new String[] { "\u0394U = ", "Q" }, true, false, base,
				"first_law_q_rearranged.gif");
		first_law_q.addRearrangedEquation(first_law_q_rearranged);
		first_law_q_rearranged.addRearrangedEquation(first_law_q);

		first_law_q_rearranged.setParentEquation(first_law_q);

		Formula first_law_w = new Formula("FirstLaw", "first_law_w",
				new String[] { "W = ", "U" }, true, false, base,
				"first_law_w.gif");
		Formula first_law_w_rearranged = new Formula("FirstLaw", 
				"first_law_w_rearranged",
				new String[] { "U = ", "W" }, true, false, base,
				"first_law_w_rearranged.gif");
		first_law_w.addRearrangedEquation(first_law_w_rearranged);
		first_law_w_rearranged.addRearrangedEquation(first_law_w);
		
		first_law_w_rearranged.setParentEquation(first_law_w);

		Formula first_law_qw = new Formula("FirstLaw", "first_law_qw",
				new String[] { "U = ", "Q", "+", "W" }, true, false, base,
				"first_law_qw.gif");
		Formula first_law_qw_rearranged_q = new Formula("FirstLaw", 
				"first_law_qw_rearranged_q", new String[] {
						"Q = ", "U", "-", "W" }, true, false, base,
				"first_law_qw_rearranged_q.gif");
		Formula first_law_qw_rearranged_w = new Formula("FirstLaw", 
				"first_law_qw_rearranged_w", new String[] {
						"W = ", "U", "-", "Q" }, true, false, base,
				"first_law_qw_rearranged_w.gif");
		first_law_qw.addRearrangedEquation(first_law_qw_rearranged_q);
		first_law_qw.addRearrangedEquation(first_law_qw_rearranged_w);
		first_law_qw_rearranged_q.addRearrangedEquation(first_law_qw);
		first_law_qw_rearranged_q
				.addRearrangedEquation(first_law_qw_rearranged_w);
		first_law_qw_rearranged_w.addRearrangedEquation(first_law_qw);
		first_law_qw_rearranged_w
				.addRearrangedEquation(first_law_qw_rearranged_q);
		
		first_law_qw_rearranged_q.setParentEquation(first_law_qw);
		first_law_qw_rearranged_w.setParentEquation(first_law_qw);

		Formula first_law_w_q = new Formula("FirstLaw", "first_law_w_q",
				new String[] { "Q = -", "W" }, true, false, base,
				"first_law_w_q.gif");
		Formula first_law_w_q_rearranged = new Formula("FirstLaw", 
				"first_law_w_q_rearranged", new String[] { "W = -", "Q" }, true, false, base,
				"first_law_w_q_rearranged.gif");
		first_law_w_q.addRearrangedEquation(first_law_w_q_rearranged);
		first_law_w_q_rearranged.addRearrangedEquation(first_law_w_q);
		
		first_law_w_q_rearranged.setParentEquation(first_law_w_q);

		equationList.add(first_law_delta_q);
		equationList.add(first_law_delta_qw);
		equationList.add(first_law_delta_w);
		equationList.add(first_law_q);
		equationList.add(first_law_w);
		equationList.add(first_law_qw);
		equationList.add(first_law_w_q);

		Collections.shuffle(equationList);
		return equationList;
	}

	private List<Formula> createDeltaUEquationList() {
		ArrayList<Formula> equationList = new ArrayList<Formula>(); //<Equation
																		// >
		equationList.add(new Formula("DeltaUExpression", "delta_u_CvT",
				new String[] { "\u0394U = ", "Cv", "*(", "T2", "-", "T1", ")" }, true,
				false, base, "delta_u_CvT.gif"));
		equationList.add(new Formula("DeltaUExpression", "delta_u_nCpT",
				new String[] { "\u0394U = ", "n", "*", "Cp", "*(", "T2", "-", "T1", ")" },
				true, false, base, "delta_u_nCpT.gif"));
		equationList.add(new Formula("DeltaUExpression", "delta_u_nCvT",
				new String[] { "\u0394U = ", "n", "*", "Cv", "*(", "T2", "-", "T1", ")" },
				true, false, base, "delta_u_nCvT.gif"));
		Collections.shuffle(equationList);
		return equationList;
	}

	private List<Formula> createIdealGasEquationList() {
		ArrayList<Formula> equationList = new ArrayList<Formula>(); //<Equation
																		// >
		Formula ideal_gas_state_nrt = new Formula("IdealGasState", "ideal_gas_state_nrt",
				new String[] { "P * V = ", "n", "*", "R", "*",
						"T" }, true, true, base, "ideal_gas_state_nrt.gif");
		Formula ideal_gas_state_nrt_rearranged_t = new Formula("IdealGasState", 
				"ideal_gas_state_nrt_rearranged_t",
				new String[] { "T = (", "P", "*", "V", ") / (", "n", "*", "R", ")"}, true, false, base,
				"ideal_gas_state_nrt_rearranged_t.gif");
		Formula ideal_gas_state_nrt_rearranged_p = new Formula("IdealGasState", 
				"ideal_gas_state_nrt_rearranged_p",
				new String[] { "P = ", "n", "*", "R", "*", "T", " / ", "V" },
				true, false, base, "ideal_gas_state_nrt_rearranged_p.gif");
		Formula ideal_gas_state_nrt_rearranged_v = new Formula("IdealGasState", 
				"ideal_gas_state_nrt_rearranged_v",
				new String[] { "V = ", "n", "*", "R", "*", "T", " / ", "P" },
				true, false, base, "ideal_gas_state_nrt_rearranged_v.gif");
		ideal_gas_state_nrt
				.addRearrangedEquation(ideal_gas_state_nrt_rearranged_t);
		ideal_gas_state_nrt
				.addRearrangedEquation(ideal_gas_state_nrt_rearranged_p);
		ideal_gas_state_nrt
				.addRearrangedEquation(ideal_gas_state_nrt_rearranged_v);
		ideal_gas_state_nrt_rearranged_t
				.addRearrangedEquation(ideal_gas_state_nrt_rearranged_v);
		ideal_gas_state_nrt_rearranged_t
				.addRearrangedEquation(ideal_gas_state_nrt_rearranged_p);
		ideal_gas_state_nrt_rearranged_p
				.addRearrangedEquation(ideal_gas_state_nrt_rearranged_v);
		ideal_gas_state_nrt_rearranged_p
				.addRearrangedEquation(ideal_gas_state_nrt_rearranged_t);
		ideal_gas_state_nrt_rearranged_v
				.addRearrangedEquation(ideal_gas_state_nrt_rearranged_p);
		ideal_gas_state_nrt_rearranged_v
				.addRearrangedEquation(ideal_gas_state_nrt_rearranged_t);
		
		ideal_gas_state_nrt_rearranged_t.setParentEquation(ideal_gas_state_nrt);
		ideal_gas_state_nrt_rearranged_p.setParentEquation(ideal_gas_state_nrt);
		ideal_gas_state_nrt_rearranged_v.setParentEquation(ideal_gas_state_nrt);

		Formula ideal_gas_state_rt = new Formula("IdealGasState", "ideal_gas_state_rt",
				new String[] { "P * V = ", "R", "*", "T" }, true,
				true, base, "ideal_gas_state_rt.gif");
		Formula ideal_gas_state_rt_rearranged_t = new Formula("IdealGasState", 
				"ideal_gas_state_rt_rearranged_t", new String[] {
						"T = ", "P", "*", "V", " / ", "R" }, true, false,
				base, "ideal_gas_state_rt_rearranged_t.gif");
		Formula ideal_gas_state_rt_rearranged_p = new Formula("IdealGasState", 
				"ideal_gas_state_rt_rearranged_p", new String[] {
						"P = ", "R", "*", "T", " / ", "V" }, true, false,
				base, "ideal_gas_state_rt_rearranged_p.gif");
		Formula ideal_gas_state_rt_rearranged_v = new Formula("IdealGasState", 
				"ideal_gas_state_rt_rearranged_v", new String[] {
						"V = ", "R", "*", "T", " / ", "P" }, true, false,
				base, "ideal_gas_state_rt_rearranged_v.gif");
		ideal_gas_state_rt
				.addRearrangedEquation(ideal_gas_state_rt_rearranged_t);
		ideal_gas_state_rt
				.addRearrangedEquation(ideal_gas_state_rt_rearranged_p);
		ideal_gas_state_rt
				.addRearrangedEquation(ideal_gas_state_rt_rearranged_v);
		ideal_gas_state_rt_rearranged_t
				.addRearrangedEquation(ideal_gas_state_rt_rearranged_v);
		ideal_gas_state_rt_rearranged_t
				.addRearrangedEquation(ideal_gas_state_rt_rearranged_p);
		ideal_gas_state_rt_rearranged_p
				.addRearrangedEquation(ideal_gas_state_rt_rearranged_v);
		ideal_gas_state_rt_rearranged_p
				.addRearrangedEquation(ideal_gas_state_rt_rearranged_t);
		ideal_gas_state_rt_rearranged_v
				.addRearrangedEquation(ideal_gas_state_rt_rearranged_p);
		ideal_gas_state_rt_rearranged_v
				.addRearrangedEquation(ideal_gas_state_rt_rearranged_t);
		
		ideal_gas_state_rt_rearranged_t.setParentEquation(ideal_gas_state_rt);
		ideal_gas_state_rt_rearranged_p.setParentEquation(ideal_gas_state_rt);
		ideal_gas_state_rt_rearranged_v.setParentEquation(ideal_gas_state_rt);

		Formula ideal_gas_state_tp = new Formula("IdealGasState", "ideal_gas_state_tp",
				new String[] {
						"P2 / P1 = ", "T2", "/", "T1"},
				true, true, base, "ideal_gas_state_tp.gif");
		Formula ideal_gas_state_tp_rearranged_p1 = new Formula("IdealGasState", 
				"ideal_gas_state_tp_rearranged_p1",
				new String[] { "P1 = ", "P2", "* ", "T1", "/", "T2"},
				true, false, base, "ideal_gas_state_tp_rearranged_p1.gif");
		Formula ideal_gas_state_tp_rearranged_p2 = new Formula("IdealGasState", 
				"ideal_gas_state_tp_rearranged_p2",
				new String[] { "P2 = ", "P1", "* ", "T2", "/", "T1"},
				true, false, base, "ideal_gas_state_tp_rearranged_p2.gif");
		Formula ideal_gas_state_tp_rearranged_t1 = new Formula("IdealGasState", "ideal_gas_state_tp_rearranged_t1",
				new String[] { "T1 = ", "P1", "/", "P2", "*", "T2"}, true, false,
				base, "ideal_gas_state_tp_rearranged_t1.gif");
		Formula ideal_gas_state_tp_rearranged_t2 = new Formula("IdealGasState", "ideal_gas_state_tp_rearranged_t2",
				new String[] { "T2 = ", "P2", "/", "P1", "*", "T1"}, true, false,
				base, "ideal_gas_state_tp_rearranged_t2.gif");
		ideal_gas_state_tp.addRearrangedEquation(ideal_gas_state_tp_rearranged_p1);
		ideal_gas_state_tp.addRearrangedEquation(ideal_gas_state_tp_rearranged_p2);
		ideal_gas_state_tp.addRearrangedEquation(ideal_gas_state_tp_rearranged_t1);
		ideal_gas_state_tp.addRearrangedEquation(ideal_gas_state_tp_rearranged_t2);
		
		ideal_gas_state_tp_rearranged_p1.addRearrangedEquation(ideal_gas_state_tp_rearranged_p2);
		ideal_gas_state_tp_rearranged_p1.addRearrangedEquation(ideal_gas_state_tp_rearranged_t1);
		ideal_gas_state_tp_rearranged_p1.addRearrangedEquation(ideal_gas_state_tp_rearranged_t2);
		
		ideal_gas_state_tp_rearranged_p2.addRearrangedEquation(ideal_gas_state_tp_rearranged_p1);
		ideal_gas_state_tp_rearranged_p2.addRearrangedEquation(ideal_gas_state_tp_rearranged_t1);
		ideal_gas_state_tp_rearranged_p2.addRearrangedEquation(ideal_gas_state_tp_rearranged_t2);
		
		ideal_gas_state_tp_rearranged_t1.addRearrangedEquation(ideal_gas_state_tp_rearranged_p1);
		ideal_gas_state_tp_rearranged_t1.addRearrangedEquation(ideal_gas_state_tp_rearranged_p2);
		ideal_gas_state_tp_rearranged_t1.addRearrangedEquation(ideal_gas_state_tp_rearranged_t2);
		
		ideal_gas_state_tp_rearranged_t2.addRearrangedEquation(ideal_gas_state_tp_rearranged_p1);
		ideal_gas_state_tp_rearranged_t2.addRearrangedEquation(ideal_gas_state_tp_rearranged_p2);
		ideal_gas_state_tp_rearranged_t2.addRearrangedEquation(ideal_gas_state_tp_rearranged_t1);
		
		ideal_gas_state_tp_rearranged_p1.setParentEquation(ideal_gas_state_tp);
		ideal_gas_state_tp_rearranged_p2.setParentEquation(ideal_gas_state_tp);
		ideal_gas_state_tp_rearranged_t1.setParentEquation(ideal_gas_state_tp);
		ideal_gas_state_tp_rearranged_t2.setParentEquation(ideal_gas_state_tp);
		
		equationList.add(ideal_gas_state_nrt);
		equationList.add(ideal_gas_state_rt);
		equationList.add(ideal_gas_state_tp);
		Collections.shuffle(equationList);
		return equationList;
	}

	private List<Formula> createAdiabaticEquationList() {
		ArrayList<Formula> equationList = new ArrayList<Formula>(); //<Equation
																		// >
		Formula adiabatic_temp_rat_1 = new Formula("Adiabatic", "adiabatic_temp_rat_1",
				new String[] { "T2 / T1 = 1" }, true, false, base,
				"adiabatic_temp_rat_1.gif");
		/*Formula adiabatic_temp_rat_P = new Formula("Adiabatic", "adiabatic_temp_rat_P",
				new String[] { "T2 / T1 = ", "P2", "/", "P1" }, true, true,
				base, "adiabatic_temp_rat_P.gif");
		Formula adiabatic_temp_rat_P_rearranged_P1 = new Formula("Adiabatic", "adiabatic_temp_rat_P_rearranged_P1",
				new String[] { "P1 = ", "T1", "/", "T2", "*", "P2"}, true, false,
				base, "adiabatic_temp_rat_P_rearranged_P1.gif");
		Formula adiabatic_temp_rat_P_rearranged_P2 = new Formula("Adiabatic", "adiabatic_temp_rat_P_rearranged_P2",
				new String[] { "P2 = ", "T2", "/", "T1", "*", "P1"}, true, false,
				base, "adiabatic_temp_rat_P_rearranged_P2.gif");
		Formula adiabatic_temp_rat_P_rearranged_T1 = new Formula("Adiabatic", "adiabatic_temp_rat_P_rearranged_T1",
				new String[] { "T1 = ", "P1", "/", "P2", "*", "T2"}, true, false,
				base, "adiabatic_temp_rat_P_rearranged_T1.gif");
		Formula adiabatic_temp_rat_P_rearranged_T2 = new Formula("Adiabatic", "adiabatic_temp_rat_P_rearranged_T2",
				new String[] { "T2 = ", "P2", "/", "P1", "*", "T1"}, true, false,
				base, "adiabatic_temp_rat_P_rearranged_T2.gif");
		adiabatic_temp_rat_P.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_P1);
		adiabatic_temp_rat_P.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_P2);
		adiabatic_temp_rat_P.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_T1);
		adiabatic_temp_rat_P.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_T2);
		adiabatic_temp_rat_P_rearranged_P1.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_P2);
		adiabatic_temp_rat_P_rearranged_P1.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_T1);
		adiabatic_temp_rat_P_rearranged_P1.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_T2);
		adiabatic_temp_rat_P_rearranged_P2.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_P1);
		adiabatic_temp_rat_P_rearranged_P2.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_T1);
		adiabatic_temp_rat_P_rearranged_P2.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_T2);
		adiabatic_temp_rat_P_rearranged_T1.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_P1);
		adiabatic_temp_rat_P_rearranged_T1.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_P2);
		adiabatic_temp_rat_P_rearranged_T1.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_T2);
		adiabatic_temp_rat_P_rearranged_T2.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_P1);
		adiabatic_temp_rat_P_rearranged_T2.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_P2);
		adiabatic_temp_rat_P_rearranged_T2.addRearrangedEquation(adiabatic_temp_rat_P_rearranged_T1);
		adiabatic_temp_rat_P_rearranged_P1.setParentEquation(adiabatic_temp_rat_P);
		adiabatic_temp_rat_P_rearranged_P2.setParentEquation(adiabatic_temp_rat_P);
		adiabatic_temp_rat_P_rearranged_T1.setParentEquation(adiabatic_temp_rat_P);
		adiabatic_temp_rat_P_rearranged_T2.setParentEquation(adiabatic_temp_rat_P);*/

		Formula adiabatic_temp_rat_P_r = new Formula("Adiabatic", 
				"adiabatic_temp_rat_P_r",
				new String[] {
						"T2 / T1 = (", "P2", "/", "P1", ")^((", "\u03B3",
						"-1)/", "\u03B3", ")" }, true, true, base,
				"adiabatic_temp_rat_P_r.gif");
		Formula adiabatic_temp_rat_rearranged_t1 = new Formula("Adiabatic", 
				"adiabatic_temp_rat_rearranged_t1",
				new String[] {
						"T1 = ", "T2", "/((", "P2", "/", "P1", ")^((",
						"\u03B3", "-1)/", "\u03B3", "))" }, true, false, base,
				"adiabatic_temp_rat_rearranged_t1.gif");
		Formula adiabatic_temp_rat_rearranged_t2 = new Formula("Adiabatic", 
				"adiabatic_temp_rat_rearranged_t2",
				new String[] { "T2 = ",
						"T1", "*(", "P2", "/", "P1", ")^((", "\u03B3", "-1)/",
						"\u03B3", ")" }, true, false, base,
				"adiabatic_temp_rat_rearranged_t2.gif");
		adiabatic_temp_rat_P_r
				.addRearrangedEquation(adiabatic_temp_rat_rearranged_t1);
		adiabatic_temp_rat_P_r
				.addRearrangedEquation(adiabatic_temp_rat_rearranged_t2);
		adiabatic_temp_rat_rearranged_t2
				.addRearrangedEquation(adiabatic_temp_rat_rearranged_t1);
		adiabatic_temp_rat_rearranged_t1
				.addRearrangedEquation(adiabatic_temp_rat_rearranged_t2);
		
		adiabatic_temp_rat_rearranged_t1.setParentEquation(adiabatic_temp_rat_P_r);
		adiabatic_temp_rat_rearranged_t2.setParentEquation(adiabatic_temp_rat_P_r);

		Formula adiabatic_temp_rat_V = new Formula("Adiabatic", "adiabatic_temp_rat_V",
				new String[] { "T2 / T1 = ", "V2", "/", "V1" }, true, false,
				base, "adiabatic_temp_rat_V.gif");

		Formula adiabatic_equation = new Formula("Adiabatic", "adiabatic_equation",
				new String[] { "P2 / P1 = (", "V1",
						"/", "V2", ")^", "\u03B3" }, true, true, base,
				"adiabatic_equation.gif");
		Formula adiabatic_equation_rearranged_v1 = new Formula("Adiabatic", 
				"adiabatic_equation_rearranged_v1",
				new String[] { "V1 = (", "P2",
						"/", "P1", ")^(1 / ", "\u03B3", ")/", "V2" }, true,
				false, base, "adiabatic_equation_rearranged_v1.gif");
		Formula adiabatic_equation_rearranged_v2 = new Formula("Adiabatic", 
				"adiabatic_equation_rearranged_v2",
				new String[] { "V2 = ", "V1",
						"/(", "P2", "/", "P1", ")^(1/", "\u03B3", ")" }, true,
				false, base, "adiabatic_equation_rearranged_v2.gif");
		Formula adiabatic_equation_rearranged_p1 = new Formula("Adiabatic", 
				"adiabatic_equation_rearranged_p1",
				new String[] { "P1 = (", "V1", "/", "V2", ")^", "\u03B3", "/",
						"P2" }, true, false, base,
				"adiabatic_equation_rearranged_p1.gif");
		Formula adiabatic_equation_rearranged_p2 = new Formula("Adiabatic", 
				"adiabatic_equation_rearranged_p2",
				new String[] { "P2 = ", "P1", "/(", "V1", "/", "V2", ")^",
						"\u03B3" }, true, false, base,
				"adiabatic_equation_rearranged_p2.gif");
		adiabatic_equation.addRearrangedEquation(adiabatic_equation_rearranged_v1);
		adiabatic_equation.addRearrangedEquation(adiabatic_equation_rearranged_v2);
		adiabatic_equation.addRearrangedEquation(adiabatic_equation_rearranged_p1);
		adiabatic_equation.addRearrangedEquation(adiabatic_equation_rearranged_p2);
		adiabatic_equation_rearranged_v1.addRearrangedEquation(adiabatic_equation_rearranged_v2);
		adiabatic_equation_rearranged_v1.addRearrangedEquation(adiabatic_equation_rearranged_p1);
		adiabatic_equation_rearranged_v1.addRearrangedEquation(adiabatic_equation_rearranged_p2);
		adiabatic_equation_rearranged_v2.addRearrangedEquation(adiabatic_equation_rearranged_v1);
		adiabatic_equation_rearranged_v2.addRearrangedEquation(adiabatic_equation_rearranged_p1);
		adiabatic_equation_rearranged_v2.addRearrangedEquation(adiabatic_equation_rearranged_p2);
		adiabatic_equation_rearranged_p1.addRearrangedEquation(adiabatic_equation_rearranged_v1);
		adiabatic_equation_rearranged_p1.addRearrangedEquation(adiabatic_equation_rearranged_p2);
		adiabatic_equation_rearranged_p1.addRearrangedEquation(adiabatic_equation_rearranged_v2);
		adiabatic_equation_rearranged_p2.addRearrangedEquation(adiabatic_equation_rearranged_v1);
		adiabatic_equation_rearranged_p2.addRearrangedEquation(adiabatic_equation_rearranged_p1);
		adiabatic_equation_rearranged_p2.addRearrangedEquation(adiabatic_equation_rearranged_v2);
		adiabatic_equation_rearranged_v1.setParentEquation(adiabatic_equation);
		adiabatic_equation_rearranged_v2.setParentEquation(adiabatic_equation);
		adiabatic_equation_rearranged_p1.setParentEquation(adiabatic_equation);
		adiabatic_equation_rearranged_p2.setParentEquation(adiabatic_equation);

		equationList.add(adiabatic_temp_rat_1);
		equationList.add(adiabatic_temp_rat_P_r);
		equationList.add(adiabatic_temp_rat_V);
		equationList.add(adiabatic_equation);
		Collections.shuffle(equationList);
		return equationList;
	}

	private List<Formula> createIsothermalWorkEquationList() {
		ArrayList<Formula> equationList = new ArrayList<Formula>(); //<Equation
																		// >
		equationList.add(new Formula("IsothermalWork", "isothermal_work_nrt_lnP",
				new String[] { "W = ", "n", "*", "R",
						"*", "T", "*log(", "P2", "/", "P1", ")" }, true, false,
				base, "isothermal_work_nrt_lnP.gif"));
		equationList.add(new Formula("IsothermalWork", "isothermal_work_nrt_lnV",
				new String[] { "W = ", "n", "*", "R",
						"*", "T", "*log(", "V1", "/", "V2", ")" }, true, false,
				base, "isothermal_work_nrt_lnV.gif"));
		equationList.add(new Formula("IsothermalWork", "isothermal_work_nrtr_P",
				new String[] { "W = (", "n", "*", "R", "*", "T1", ")/(",
						"\u03B3", "-1)*((", "P2", "/", "P1", ")^((", "\u03B3",
						"-1)/", "\u03B3", ")-1)" }, true, false, base,
				"isothermal_work_nrtr_P.gif"));
		equationList.add(new Formula("IsothermalWork", "isothermal_work_rt_lnP",
				new String[] { "W = ", "R", "*", "T",
						"*log(", "P2", "/", "P1", ")" }, true, false, base,
				"isothermal_work_rt_lnP.gif"));
		equationList.add(new Formula("IsothermalWork", "isothermal_work_rt_lnV",
				new String[] { "W = ", "R", "*", "T",
						"*log(", "V1", "/", "V2", ")" }, true, false, base,
				"isothermal_work_rt_lnV.gif"));
		equationList.add(new Formula("IsothermalWork", "isothermal_work_rtr_P",
				new String[] { "W = (", "R", "*", "T1", ")/(", "\u03B3",
						"-1)*((", "P2", "/", "P1", ")^((", "\u03B3", "-1)/",
						"\u03B3", ")-1)" }, true, false, base,
				"isothermal_work_rtr_P.gif"));
		Collections.shuffle(equationList);
		return equationList;
	}

	private List<Formula> createSpecialPropertiesEquationList() {
		ArrayList<Formula> equationList = new ArrayList<Formula>();
		equationList.add(new Formula("TransitionSpecialProperty", "special_properties_q_0",
				new String[] { "Q = 0" }, false, false, base,
				"special_properties_q_0.gif"));
		equationList.add(new Formula("TransitionSpecialProperty", "special_properties_w_0",
				new String[] { "W = 0" }, false, false, base,
				"special_properties_w_0.gif"));
		equationList.add(new Formula("TransitionSpecialProperty", "special_properties_u_0",
				new String[] { "\u0394U = 0" }, false, false, base,
				"special_properties_u_0.gif"));
		equationList.add(new Formula("StateSpecialProperty", "special_properties_v2_v1",
				new String[] { "V2 = ", "V1" }, true, false, base,
				"special_properties_v2_v1.gif"));
		equationList.add(new Formula("StateSpecialProperty", "special_properties_t2_t1",
				new String[] { "T2 = ", "T1" }, true, false, base,
				"special_properties_t2_t1.gif"));
		equationList.add(new Formula("StateSpecialProperty", "special_properties_p2_p1",
				new String[] { "P2 = ", "P1" }, true, false, base,
				"special_properties_p2_p1.gif"));
		Collections.shuffle(equationList);
		return equationList;
	}

	private List<Formula> createIsobaricWorkEquationList() {
		ArrayList<Formula> equationList = new ArrayList<Formula>();
		equationList.add(new Formula("IsobaricWork", "isobaric_work_pt",
				new String[] { "W = -", "P", "*(", "T2", "-", "T1", ")" },
				true, false, base, "isobaric_work_pt.gif"));
		equationList.add(new Formula("IsobaricWork", "isobaric_work_v",
				new String[] { "W = -", "P", "*(", "V2", "-", "V1", ")" },
				true, false, base, "isobaric_work_v.gif"));
		equationList.add(new Formula("IsobaricWork", "isobaric_work_t",
				new String[] { "W = -", "R", "*(", "T2", "-", "T1", ")" },
				true, false, base, "isobaric_work_t.gif"));
		Collections.shuffle(equationList);
		return equationList;
	}

	private List<SpecialValue> createSpecialValueCvList() {
		ArrayList<SpecialValue> specialValues = new ArrayList<SpecialValue>();
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_CV,
				12.471, subject.getNextID(), base, "special_value_cv_32.gif"));
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_CV,
				20.785, subject.getNextID(), base, "special_value_cv_52.gif"));
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_CV,
				29.099, subject.getNextID(), base, "special_value_cv_72.gif"));
		Collections.shuffle(specialValues);
		return specialValues;
	}

	private List<SpecialValue> createSpecialValueCpList() {
		ArrayList<SpecialValue> specialValues = new ArrayList<SpecialValue>();
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_CP,
				12.471, subject.getNextID(), base, "special_value_cp_32.gif"));
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_CP,
				20.785, subject.getNextID(), base, "special_value_cp_52.gif"));
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_CP,
				29.099, subject.getNextID(), base, "special_value_cp_72.gif"));
		Collections.shuffle(specialValues);
		return specialValues;
	}

	private List<SpecialValue> createSpecialValueRList() {
		ArrayList<SpecialValue> specialValues = new ArrayList<SpecialValue>();
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_R, 1.986,
				subject.getNextID(), base, "special_value_r_1986.gif"));
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_R, 8.314,
				subject.getNextID(), base, "special_value_r_8314.gif"));
		specialValues
				.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_R, 62.364,
						subject.getNextID(), base, "special_value_r_62364.gif"));
		Collections.shuffle(specialValues);
		return specialValues;
	}

	private List<SpecialValue> createSpecialValueGammaList() {
		ArrayList<SpecialValue> specialValues = new ArrayList<SpecialValue>();
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_GAMMA_WORD,
				1.4, subject.getNextID(), base, "special_value_y_14.gif"));
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_GAMMA_WORD,
				1.667, subject.getNextID(), base, "special_value_y_1667.gif"));
		Collections.shuffle(specialValues);
		return specialValues;
	}
	
	private List<SpecialValue> createSpecialValueNList() {
		ArrayList<SpecialValue> specialValues = new ArrayList<SpecialValue>();
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_N, 1,
				subject.getNextID(), base, "special_value_n_1.gif"));
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_N, 2,
				subject.getNextID(), base, "special_value_n_2.gif"));
		specialValues.add(new SpecialValue(SpecialValue.SPECIAL_VALUE_N, 50,
				subject.getNextID(), base, "special_value_n_50.gif"));
		Collections.shuffle(specialValues);
		return specialValues;
	}
	
	/*public static void main(String[] args)
	{
		System.out.println("here");
		ProcessApplet me = new ProcessApplet();
		//System.out.println("Adiabatic");
		for (Formula formula : me.createAdiabaticEquationList()) {
			System.out.println(formula.formulaForASPIRE() + " :: " + formula.name());
			for (Formula rearranged : formula.getRearrangedEquations()) {
				System.out.println(rearranged.formulaForASPIRE() + " :: " + rearranged.name());
			}
		}
		//System.out.println("Delta U");
		for (Formula formula : me.createDeltaUEquationList()) {
			System.out.println(formula.formulaForASPIRE() + " :: " + formula.name());
			for (Formula rearranged : formula.getRearrangedEquations()) {
				System.out.println(rearranged.formulaForASPIRE() + " :: " + rearranged.name());
			}
		}
		//System.out.println("First Law");
		for (Formula formula : me.createFirstLawEquationList()) {
			System.out.println(formula.formulaForASPIRE() + " :: " + formula.name());
			for (Formula rearranged : formula.getRearrangedEquations()) {
				System.out.println(rearranged.formulaForASPIRE() + " :: " + rearranged.name());
			}
		}
		//System.out.println("Ideal Gas");
		for (Formula formula : me.createIdealGasEquationList()) {
			System.out.println(formula.formulaForASPIRE() + " :: " + formula.name());
			for (Formula rearranged : formula.getRearrangedEquations()) {
				System.out.println(rearranged.formulaForASPIRE() + " :: " + rearranged.name());
			}
		}
		//System.out.println("Isobaric Work");
		for (Formula formula : me.createIsobaricWorkEquationList()) {
			System.out.println(formula.formulaForASPIRE() + " :: " + formula.name());
			for (Formula rearranged : formula.getRearrangedEquations()) {
				System.out.println(rearranged.formulaForASPIRE() + " :: " + rearranged.name());
			}
		}
		//System.out.println("Isothermal Work");
		for (Formula formula : me.createIsothermalWorkEquationList()) {
			System.out.println(formula.formulaForASPIRE() + " :: " + formula.name());
			for (Formula rearranged : formula.getRearrangedEquations()) {
				System.out.println(rearranged.formulaForASPIRE() + " :: " + rearranged.name());
			}
		}
		//System.out.println("Special Properties");
		for (Formula formula : me.createSpecialPropertiesEquationList()) {
			System.out.println(formula.formulaForASPIRE() + " :: " + formula.name());
			for (Formula rearranged : formula.getRearrangedEquations()) {
				System.out.println(rearranged.formulaForASPIRE() + " :: " + rearranged.name());
			}
		}
	}*/
}
